/* 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is the Netscape Portable Runtime library.
 * 
 * The Initial Developer of the Original Code is Netscape
 * Communications Corporation.  Portions created by Netscape are 
 * Copyright (C) 1994-2000 Netscape Communications Corporation.  All
 * Rights Reserved.
 * 
 * Contributor(s):  Silicon Graphics, Inc.
 * 
 * Portions created by SGI are Copyright (C) 2000-2001 Silicon
 * Graphics, Inc.  All Rights Reserved.
 * 
 * Alternatively, the contents of this file may be used under the
 * terms of the GNU General Public License Version 2 or later (the
 * "GPL"), in which case the provisions of the GPL are applicable 
 * instead of those above.  If you wish to allow use of your 
 * version of this file only under the terms of the GPL and not to
 * allow others to use your version of this file under the MPL,
 * indicate your decision by deleting the provisions above and
 * replace them with the notice and other provisions required by
 * the GPL.  If you do not delete the provisions above, a recipient
 * may use your version of this file under either the MPL or the
 * GPL.
 */

/*
 * This file is derived directly from the State Threads library, 
 * modified extensively by P.E. Hadjidoukas for PSthreads/UthLib. 
 */

#include "uth_common.h"
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#if !defined(WIN32_MSVC) && !defined(MINGW)
#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>
#include <time.h>
#else
/*#define _WIN32_WINNT	0x0500*/
#include <windows.h>
#endif
#include <string.h>

/* Internal global data - defined in common.h */

struct uth_data	*_uth_data;
uth_stats_t			_uth_stats;


/*
 * Queue management for thread recycling
 */

/* Queue Initialization */
void _uth_queue_init (queue_t * q)
{
	_queue_init(q);
}

/* Insert a thread into the global reuse queue */
void _uth_to_reuseq (uth_t *nth)
{
	_enqueue_head(&reuse_uth_q, nth);
}

/* Extract a thread from the global reuse queue */
uth_t *_uth_reuseq_dequeue ()
{
	uth_t *nth = NULL;
    
	_dequeue(&reuse_uth_q, &nth);
	return nth;
}

/* Insert a thread into a local reuse queue */
void _uth_to_reuselrq (int qid, register uth_t *nth)
{
	_enqueue_head_bare(&reuse_uth_lq[qid], nth);
}

/* Extract a thread from a local reuse queue */
uth_t *_uth_reuselrq_dequeue (int qid)
{
	uth_t *nth = NULL;
    
    	_dequeue_bare(&reuse_uth_lq[qid], &nth);
	return nth;
}

/* General routine: returns a recycled thread */
uth_t *_uth_get_reused_thread(void)
{
	uth_t *thread;

#if defined(UTH_LOCAL_REUSE)
	int vp = uth_get_vpid();
	thread = _uth_reuselrq_dequeue (vp);
#else
	thread = _uth_reuseq_dequeue();
#endif

	return thread;
}

/* General routine: recycles a finished thread */
void _uth_put_reused_thread(uth_t *thread)
{
#if defined(UTH_LOCAL_REUSE)
	int vp = uth_get_vpid();
/*	int vp = thread->vp_id;	*/

	_uth_to_reuselrq(vp, thread);
#else
	_uth_to_reuseq(thread);
#endif
}

/* 
 * Get system page size (syscall not available on Win32 and QNX platforms)
 */ 

#if defined(WIN32_MSVC)
size_t getpagesize(void)
{
	SYSTEM_INFO sysinf;
	GetSystemInfo(&sysinf);
	return (sysinf.dwPageSize);
}
#elif defined(QNX)
size_t getpagesize(void)
{
	return 4096;
}
#endif

void _uth_data_init ()
{
	char *mem;

	mem = _ort_calloc_aligned(sizeof (struct uth_data), NULL);
	if (mem == NULL) {
		fprintf (stderr, "Allocating uth_data (size %ld): malloc failed",
				(unsigned long)sizeof (struct uth_data));
		exit (1);
	}

	_uth_data = (void *) mem;
}

void print_uth_data ()
{
	int i;
	printf ("print_uth_data          0x%lx\n", (unsigned long) print_uth_data);

	printf ("uth_data at             0x%lx\n", (unsigned long) _uth_data);
	printf ("sizeof (uth_data)       0x%lx (%ld)\n", (unsigned long) sizeof(_uth_data), (unsigned long)sizeof(_uth_data));

	printf (" reuse_q                0x%lx\n", (unsigned long) &reuse_uth_q);
	printf (" reuse_lq               0x%lx\n", (unsigned long) &reuse_uth_lq);
	i = 0;
	while (i<MAX_VPS) {
		printf (" reuse_lq[%2d]          0x%lx\n", i, (unsigned long) &reuse_uth_lq[i]);
		++i;
	}

	printf (" _UTH_CURRENT_THREAD    0x%lx\n", (unsigned long) &_UTH_CURRENT_THREAD(0));
	printf (" _UTH_VP_IDS            0x%lx\n", (unsigned long) &vp_ids);

	printf (" _UTH_PAGE_SIZE         0x%lx\n", (unsigned long) &_UTH_PAGE_SIZE);
	printf (" _UTH_STACK_SIZE        0x%lx\n", (unsigned long) &_UTH_STACK_SIZE);
}

/* General library initialization */
static int initialized = 0;

int uth_init(long stacksize)
{
	int i;

	if (initialized)
		return 0;
 
	_uth_data_init();
	if ((unsigned long) _uth_data & (CACHE_LINE_SIZE-1)) {
/*		fprintf (stderr, "WARNING: uth_data structure not cacheline-aligned (0x%lx)\n",
				(unsigned long) _uth_data);
*/
	}
#if 0
	print_uth_data();
#endif
	_uth_queue_init(&reuse_uth_q);
	for (i = 0; i < MAX_VPS; i++) {
		_uth_queue_init(&reuse_uth_lq[i]);
	}
	_UTH_PAGE_SIZE = getpagesize();
	if (stacksize == 0)
		_UTH_STACK_SIZE = UTH_DEFAULT_STACK_SIZE;
	else
		_UTH_STACK_SIZE = stacksize;

	memset(&_uth_stats, 0, sizeof(uth_stats_t));

	initialized = 1;
	return 0;
}

/* 
 * Initialize the current virtual processor (Posix Thread) and assign
 * an identifier (vpid) to it
 */
uth_t *uth_vp_init(int vpid, void *arg)
{
	uth_t *thread;
	int i;

	if (!initialized) {
		_uth_data_init();
		_uth_queue_init(&reuse_uth_q);
		for (i = 0; i < MAX_VPS; i++) {
			_uth_queue_init(&reuse_uth_lq[i]);
		}
		_UTH_PAGE_SIZE = getpagesize();
		_UTH_STACK_SIZE = UTH_DEFAULT_STACK_SIZE;

		memset(&_uth_stats, 0, sizeof(uth_stats_t));

		initialized = 1;
	}

	vp_ids[vpid][0] = pthread_self();	/* store my pthread_t - used for self identification */


	thread = (uth_t *) calloc(1, sizeof(uth_t));	/* associate a user level thread with this Posix Thread */
	if (!thread)
		return (void *)-1;

	thread->vp_id = vpid;
	thread->arg = NULL;

	_UTH_SET_CURRENT_THREAD(vpid, thread);

	return thread;
}

/* 
 * Warning: if we use stack-based self identification,
 * this function must be called only once, during the initialiation,
 * (just after uth_init(0)) 
 */
int uth_set_stacksize(int stacksize)
{
	int previous = _UTH_STACK_SIZE;
	_UTH_STACK_SIZE = stacksize;
	if (stacksize == 0) _UTH_STACK_SIZE = UTH_DEFAULT_STACK_SIZE;

	return previous;
}

/* Get the identifier (vpid) of the current Posix Thread */
int uth_get_vpid(void)
{
	int i;

	pthread_t me = pthread_self();

	for (i = 0; i < MAX_VPS; i++)
		if (pthread_equal(me, vp_ids[i][0])) return i;
		
	printf("fatal error\n");
	exit(1);

	/* Not reached */
	return -1;
}

/* Driver routine */
void _uth_main(void)
{ 
	uth_t *thread;
	int vp  = uth_get_vpid();

	thread = _UTH_CURRENT_THREAD(vp);

	/* Run thread main */
	(*thread->start)(thread->arg);
	/* I can put a cleanup routine here */
}

/* 
 * Creates an (underlying) thread. The thread descriptor
 * is allocated in the stack of the thread
 */
uth_t *uth_create(void (*start)(void *arg), void *arg)
{
	uth_t *thread;
	uth_stack_t *stack;
	uth_stack_t mystack;
	int stk_size;
	char *sp;
#ifdef __ia64__
	char *bsp;
#endif
	stk_size = _UTH_STACK_SIZE;	/* could be function argument */
#if defined(UTH_REUSE)
	/* Try to reuse a finished thread */
	if ((thread = _uth_get_reused_thread()) == NULL) {
#endif
		/* Thread/stack allocation and initialization */

		/* Adjust stack size */
		stk_size = ((stk_size + _UTH_PAGE_SIZE - 1) / _UTH_PAGE_SIZE) * _UTH_PAGE_SIZE;
		stack = _uth_stack_new(&mystack, stk_size);
		if (!stack)
			return NULL;

		/* Allocate thread object and per-thread data off the stack */
#if defined(MD_STACK_GROWS_DOWN)
		sp = stack->stk_top;
#ifdef __ia64__
		/*
		 * The stack segment is split in the middle. The upper half is used
		 * as backing store for the register stack which grows upward.
		 * The lower half is used for the traditional memory stack which
		 * grows downward. Both stacks start in the middle and grow outward
		 * from each other.
		 */
		sp -= (stk_size >> 1);
		bsp = sp;
		/* Make register stack 64-byte aligned */
		if ((unsigned long)bsp & 0x3f)
			bsp = bsp + (0x40 - ((unsigned long)bsp & 0x3f));
		stack->bsp = bsp + _UTH_STACK_PAD_SIZE;
#endif
		/* thread descriptor is stored in its stack */
		sp = sp - sizeof(uth_t);
		thread = (uth_t *) sp;

#if !defined(UTH_PAGE_ALIGNED_STACK)
		/* Make stack 64-byte aligned */
		if ((unsigned long)sp & 0x3f)
			sp = sp - ((unsigned long)sp & 0x3f);
		stack->sp = sp - _UTH_STACK_PAD_SIZE;
#else
		/* Make stack page aligned */
		if ((unsigned long)sp & (_UTH_PAGE_SIZE-1))
			sp = sp - ((unsigned long)sp & (_UTH_PAGE_SIZE-1));
		stack->sp = sp;	/*- _UTH_STACK_PAD_SIZE;*/
#endif
#elif defined(MD_STACK_GROWS_UP)
		sp = stack->stk_bottom;
		thread = (uth_t *) sp;
		sp = sp + sizeof(uth_t);

#ifndef PAGE_ALIGNED_STACK
		/* Make stack 64-byte aligned */
		if ((unsigned long)sp & 0x3f)
			sp = sp + (0x40 - ((unsigned long)sp & 0x3f));
		stack->sp = sp + _UTH_STACK_PAD_SIZE;
#else
		/* Make stack page-byte aligned (sparc -> 8192!!)*/
		if ((unsigned long)sp & (_UTH_PAGE_SIZE-1))
			sp = sp + ((_UTH_PAGE_SIZE) - ((unsigned long)sp & (_UTH_PAGE_SIZE-1)));
		stack->sp = sp; /*+ _UTH_STACK_PAD_SIZE;*/
#endif
#else
#error Unknown OS
#endif
		memset(thread, 0, sizeof(uth_t));

		/* Initialize thread */
		memcpy(&thread->stackinfo, stack, sizeof(uth_stack_t));
		thread->stack = &thread->stackinfo;
		thread->start = start;
		thread->arg = arg;

		thread->retval = (void *)0xDEAD;

#ifndef __ia64__
		_UTH_INIT_CONTEXT(thread, stack->sp, _uth_main);
#else
		_UTH_INIT_CONTEXT(thread, stack->sp, stack->bsp, _uth_main);
#endif

#if defined(UTH_REUSE)
	}
	else {
#if defined(UTH_STATS)
		_lock_acquire(&_uth_stats.mutex);
		_uth_stats.stacks_reused++;
		_lock_release(&_uth_stats.mutex);
#endif
		/* Recycled thread: just reinitialize it */
		uth_reinit(thread, start, arg);
	}
#endif

	return thread;
}

/* Thread reinitialization: set routine and argument, initialize context */
void uth_reinit(uth_t *thread, void (*start)(void *arg), void *arg)
{
	thread->start = start;
	thread->arg = arg;
#ifndef __ia64__
	_UTH_INIT_CONTEXT(thread, thread->stack->sp, _uth_main);
#else
	_UTH_INIT_CONTEXT(thread, thread->stack->sp, thread->stack->bsp, _uth_main);
#endif
}

void uth_restart(uth_t *thread)	
{
#ifndef __ia64__
	_UTH_INIT_CONTEXT(thread, thread->stack->sp, _uth_main);
#else
	_UTH_INIT_CONTEXT(thread, thread->stack->sp, thread->stack->bsp, _uth_main);
#endif
}

/* Delete (or recycle) thread */
void uth_delete(uth_t *thread)
{
#if defined(UTH_REUSE)
	_uth_put_reused_thread(thread);
#else
	_uth_stack_free(thread->stack);
#endif
}

/* 
 * Context switch: from old (i.e. current) to next. 
 * If old == NULL, switch to next without saving the
 * execution state of the old thread
 */
void uth_switchto(uth_t *old, uth_t *next)
{
	int vp = uth_get_vpid();

#if defined(UTH_CTXSW_SJLJ)
#if defined(WIN32_MSVC)	/* This is only for windows */
	unsigned long regs[6];
#endif
#endif

	if (next == NULL) {
		printf("uth_switchto: next == NULL\n");
		exit(1);
	}

#if defined(UTH_CTXSW_SJLJ)
#if defined(WIN32_MSVC)	/* This is only for windows */
	memcpy(regs, &next->context, sizeof(regs));
	MD_SETJMP(next->context);	/* re-init */
	memcpy(&next->context, regs, sizeof(regs));
#endif
#endif

	next->vp_id = vp;

	if (old != NULL) {
		_UTH_SWITCH_CONTEXT(vp, old, next);	/* save and switch */
	}
	else {
		_UTH_RESTORE_CONTEXT(vp, next);		/* just switch */
	}
}

/* Self identification */
uth_t *uth_self()
{
	uth_t *thread;
	int vp = uth_get_vpid();

	thread =  _UTH_CURRENT_THREAD(vp);
	return thread; 
}

/* Return thread argument: can be also used for thread local storage */
void *uth_getarg(uth_t *thread)
{
	return thread->arg;
}

/* Set thread argument */
void uth_setarg(uth_t *thread, void *arg)
{
	thread->arg = arg;
}


/* 
 * Secondary (extended) Interface:
 * Avoid calling uth_get_vpid() when the vpid of the current Posix thread
 * (virtual processor) is already known. For instance, this could happen
 * if the runtime system built on top of uthlib holds this information
 */

uth_t *uth_self_ex(int vp)
{
	return _UTH_CURRENT_THREAD(vp);
}

void *uth_getarg_ex(int vp)
{
	uth_t *thread;

	thread =  _UTH_CURRENT_THREAD(vp);
	return thread->arg;
}


void uth_switchto_ex(int vp, uth_t *prev, uth_t *next)
{
#if defined(UTH_CTXSW_SJLJ)
#if defined(WIN32_MSVC)	/* This is only for windows */
	unsigned long regs[6];
#endif
#endif
	if (next == NULL) {
		printf("uth_switchto: next == NULL\n");
		exit(1);
	}

#if defined(UTH_CTXSW_SJLJ)
#if defined(WIN32_MSVC)	/* This is only for windows */
	memcpy(regs, &next->context, sizeof(regs));
	MD_SETJMP(next->context);	/* re-init */
	memcpy(&next->context, regs, sizeof(regs));
#endif
#endif

	next->vp_id = vp;

	if (prev != NULL) {
		_UTH_SWITCH_CONTEXT(vp, prev, next);	/* save and switch */
	}
	else {
		/* 'prev' is NULL but actually available and can be passed through.
		 * The previous thread must be recycled: next->old_uth = old, and
		 * when 'next' restarts can recycle it.
		 */
		_UTH_RESTORE_CONTEXT(vp, next);		/* just switch */
	}
}


/* 
 * Return elapsed time in sec
 */

#if defined(WIN32_MSVC) || defined(MINGW)
LARGE_INTEGER TicksPerSecond = {0,0}; /* Global ticks per second */
double TicksPerMicrosecond = 0;       /* Global ticks/microsecond */

double ElapsedMicroseconds()
{
	LARGE_INTEGER StopTime;
	double delta;
	
	QueryPerformanceCounter(&StopTime); /* read current clock */
	if (TicksPerSecond.LowPart == 0) {
		QueryPerformanceFrequency(&TicksPerSecond);
		TicksPerMicrosecond = (float) (TicksPerSecond.LowPart/ 1E6);
	}
	delta = StopTime.QuadPart / TicksPerMicrosecond; /* low part */
	
	return delta;
}
#endif

double uth_gettime(void)
{
#if defined(WIN32_MSVC) || defined(MINGW)
	return ElapsedMicroseconds() / 1E6;
#else
	struct timeval t;
	gettimeofday(&t, NULL);
	return (double)t.tv_sec + (double)t.tv_usec*1.0E-6;
#endif
}


/* 
 * Sleep for ms milliseconds (kernel level)
 */
void uth_vp_sleep(int ms)
{
#if defined(WIN32_MSVC) || defined(MINGW)
	Sleep(ms);
#else
	struct timespec ts;

	ts.tv_sec = ms/1000;
	ts.tv_nsec = (ms%1000)*1E6;
	nanosleep(&ts,NULL);
#endif
}

unsigned long uth_vp_self(void)
{
#if !defined(WIN32_MSVC) && !defined(MINGW)
	return (unsigned long) pthread_self();
#else
	return (unsigned long) pthread_getw32threadhandle_np(pthread_self());
#endif
}

